CHINESE = "Simplified Chinese"
ENGLISH = "English"
JAPANESE = "Japanese"
KOREAN = "Korean"
SPANISH = "Spanish"
FRENCH = "French"
GERMAN = "German"
PORTUGUESE = "Portuguese"
TRADITIONAL_CHINESE = "Traditional Chinese"
RUSSIAN = "Russian"
ITALIAN = "Italian"
ARABIC = "Arabic"
HINDI = "Hindi"
DUTCH = "Dutch"
POLISH = "Polish"
TURKISH = "Turkish"
VIETNAMESE = "Vietnamese"
THAI = "Thai"
INDONESIAN = "Indonesian"
SWEDISH = "Swedish"
DANISH = "Danish"
NORWEGIAN = "Norwegian"
FINNISH = "Finnish"
